import argparse
import sys

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from scipy.stats import norm
from statsmodels.api import qqplot

pd.options.display.max_rows = 999
pd.options.display.max_colwidth = 100

import bokeh
from bokeh.models import ColumnDataSource, BoxAnnotation
from bokeh.models import ColumnDataSource
from bokeh.plotting import figure, output_file, save
from bokeh.palettes import Spectral6

sys.path.append("/workdir/bin/src/")
from functions.SAX import PAA_step, SAX_step
from functions.load_data import load
from functions.INT import INT_step
from functions.apriori import apriori
from functions.SBGen import SBGen

def plot_original(df, i_plot):
    """
    Plots the i_plot th time series of the input dataframe.
    """
    fig, ax = plt.subplots(figsize=(13, 5))

    ax.plot(df['t'], df.iloc[:, i_plot])
    ax.set_title(f'Original data - Column {df.columns[i_plot]}')
    ax.set_xlabel('Time')
    ax.set_ylabel(f'Column {df.columns[i_plot]}')

    return fig

def plot_INT(df, df_INT, i_plot):
    fig, ax = plt.subplots(figsize=(13, 5))
    ax.plot(df_INT['t'], df_INT.iloc[:, i_plot])
    ax.set_title(
        f'Inverse Normal Transformation - Column {df.columns[i_plot]}')
    ax.set_xlabel('Time')
    ax.set_ylabel(f'Column {df_INT.columns[i_plot]}')
    return fig

def plot_QQ(df, df_INT, i_plot):
    fig = plt.figure(figsize=(13, 5))
    ax = fig.add_subplot(1, 2, 1)
    qqplot(df.iloc[:, i_plot], line='45', ax=ax, c='royalblue', markersize=0.5)
    ax.set_xlabel('Normal Theoretical Quantiles')
    ax.set_ylabel('Data Quantiles')
    ax.set_title(f'QQ-plot - Original data\nColumn {df.columns[i_plot]}')

    ax = fig.add_subplot(1, 2, 2)
    qqplot(df_INT.iloc[:, i_plot], line='45',
           ax=ax, c='royalblue', markersize=0.5)
    ax.set_xlabel('Normal Theoretical Quantiles')
    ax.set_ylabel('Data Quantiles')
    ax.set_title(f'QQ-plot - Normalized data\nColumn {df.columns[i_plot]}')
    return fig

def plot_SAX(df_SAX, w, a, i_plot):
    """
    Plot the SAX representation of column i_plot.
    """

    # Assign an integer to each letter (for plotting purposes)
    breakpoints = norm.ppf(np.linspace(0, 1, a+1)[1:-1])
    alphabet = np.array([chr(i)
                         for i in range(97, 97 + a)])[:len(breakpoints)+1]
    dico = dict(zip([np.where(alphabet == e)[0][0]
                     for e in alphabet], alphabet))
    reverse_dico = {v: k for k, v in dico.items()}

    # Make some arrays to plot more easily
    sax = [reverse_dico[e] for e in df_SAX.iloc[:, i_plot]]

    fig = plt.figure(figsize=(13, 5))

    # Plot step function
    plt.step(df_SAX['t'], sax, where='post', c='k', linewidth=2, alpha=0.8)

    plt.title(
        f'SAX Representation - w = {w}, a = {a}\nColumn {df_SAX.columns[i_plot]}')
    plt.xticks(rotation='vertical')
    plt.yticks(range(len(alphabet)), alphabet)
    plt.gca().get_yticklabels()[0].set_color('crimson')
    plt.gca().get_yticklabels()[-1].set_color('crimson')
    plt.grid()

    return fig


def plot_PAA(df_INT, df_PAA, df_SAX, w, a, i_plot):
    """
    Plot the PAA representation of column i_plot.
    """

    dt = pd.to_timedelta(df_PAA['t'].iloc[1]-df_PAA['t'].iloc[0])/2

    # bokeh parameters
    source = ColumnDataSource(data=dict(
        t=df_PAA['t'],
        paa=df_PAA.iloc[:, i_plot],
        sax=df_SAX.iloc[:, i_plot],
        t_text=df_PAA['t']+dt,
        paa_text=df_PAA.iloc[:, i_plot]+0.01,
    ))

    breakpoints = norm.ppf(np.linspace(0, 1, a+1)[1:-1])
    alphabet = np.array([chr(i)
                         for i in range(97, 97 + a)])[:len(breakpoints)+1]

    p = figure(title=f'PAA Representation - w = {w}, a = {a}\nColumn {df_SAX.columns[i_plot]}',
               width=800, height=500, x_axis_type='datetime')

    # Color dict
    call_colors = bokeh.palettes.all_palettes['RdBu'][len(breakpoints)+1]
    color_key_value_pairs = list(zip(alphabet, call_colors))
    color_dict = dict(color_key_value_pairs)

    # Plot step function
    p.step(x='t', y='paa', source=source, line_width=3,
           line_alpha=0.5, line_color='black', mode='after')

    # If w is small enough, show letters
    if w <= 20:
        p.text(x='t_text', y='paa_text', source=source,
               text='sax', angle=0, text_color='black')

    p.line(x=[df_PAA['t'].iloc[-1], df_INT['t'].iloc[-1]],
           y=[df_PAA.iloc[-1, i_plot], df_PAA.iloc[-1, i_plot]],
           line_width=3, line_alpha=0.5, line_color='black')

    # Plot breakpoints horizontal lines
    for b in range(len(breakpoints)):
        if b == 0:
            annot = BoxAnnotation(
                top=breakpoints[b], fill_alpha=0.3, fill_color=color_dict[alphabet[b]])
        else:
            annot = BoxAnnotation(bottom=breakpoints[b-1], top=breakpoints[b],
                                  fill_alpha=0.3, fill_color=color_dict[alphabet[b]])
        p.add_layout(annot)

    annot = BoxAnnotation(
        bottom=breakpoints[b], fill_alpha=0.3, fill_color=color_dict[alphabet[b]])
    p.add_layout(annot)
    p.ygrid.ticker = breakpoints

    return p


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-w", type=int, default=20)
    parser.add_argument("-a", type=int, default=5)
    parser.add_argument("-s", "--min-supp", type=float, default=0.01)
    parser.add_argument("-c", "--min-conf", type=float, default=0.5)
    parser.add_argument("-i", "--i-plot", type=int, default=2)
    parser.add_argument("input", type=str)
    args = parser.parse_args(sys.argv[1:])
    w = args.w
    a = args.a
    min_supp = args.min_supp
    min_conf = args.min_conf
    i_plot = args.i_plot
    input = args.input

    df = load(input)
    # INT step
    df_INT = INT_step(df, c=3.0/8)

    # SAX step : PAA + discretization
    df_PAA = PAA_step(df_INT, w)
    df_SAX, symb_top, symb_bottom = SAX_step(df_INT, df_PAA, a)

    # Generation of Symbol Baskets
    B, support = SBGen(df_SAX, symb_top, symb_bottom, df_SAX.columns)

    # Association Rules Mining with the Apriori algorithm
    rules = apriori(B, min_supp=min_supp, min_conf=min_conf)

    # show results
    rules.to_html("rules.html")
    support.to_html("support.html")

    fig = plot_original(df, i_plot)
    plt.savefig("original.png", pad_inches=0)
    plt.close()

    fig = plot_INT(df, df_INT, i_plot)
    plt.savefig("int.png", pad_inches=0)
    plt.close()

    fig = plot_QQ(df, df_INT, i_plot)
    plt.savefig("qq.png", pad_inches=0)
    plt.close()

    fig = plot_SAX(df_SAX, w, a, i_plot)
    plt.savefig("sax.png", pad_inches=0)
    plt.close()

    output_file("paa.html", title="paa", mode='cdn')
    p = plot_PAA(df_INT,df_PAA, df_SAX, w, a, i_plot)
    save(p)





